@@ -8,12 +8,19 @@
 #include "printermanager.h"
 #include "pmkcm_log.h"
 
+#ifdef SCP_INSTALL
+#include "scpinstaller.h"
+#endif
+
 #include <QDBusConnection>
+#include <QDBusConnectionInterface>
+#include <QDBusMetaType>
 #include <QDBusPendingCallWatcher>
 #include <QDBusPendingReply>
 #include <QStringDecoder>
 
 #include <KLocalizedString>
+#include <KOSRelease>
 
 #include <KCupsRequest.h>
 #include <PPDModel.h>
@@ -48,20 +55,25 @@ PrinterManager::PrinterManager(QObject *parent, const KPluginMetaData &metaData)
                         {QLatin1String(CUPS_SERVER_REMOTE_ADMIN), false}})
 {
     setButtons(KQuickConfigModule::NoAdditionalButton);
+    initOSRelease();
 
     // Make sure we update our server settings if the user changes anything on
     // another interface
-    connect(KCupsConnection::global(), &KCupsConnection::serverAudit, this, [this](const QString &msg) {
-        serverEvent(u"AUDIT"_s, false, msg);
+    connect(KCupsConnection::global(), &KCupsConnection::serverAudit, this, [](const QString &msg) {
+        qCDebug(PMKCM) << "CUPS SERVER AUDIT" << msg;
     });
     connect(KCupsConnection::global(), &KCupsConnection::serverStarted, this, [this](const QString &msg) {
-        serverEvent(u"STARTED"_s, true, msg);
+        qCDebug(PMKCM) << "CUPS SERVER STARTED" << msg;
+        Q_EMIT serverStarted();
     });
     connect(KCupsConnection::global(), &KCupsConnection::serverStopped, this, [this](const QString &msg) {
-        serverEvent(u"STOPPED"_s, false, msg);
+        qCDebug(PMKCM) << "CUPS SERVER STOPPED" << msg;
+        m_serverSettingsLoaded = false;
+        Q_EMIT serverStopped();
     });
     connect(KCupsConnection::global(), &KCupsConnection::serverRestarted, this, [this](const QString &msg) {
-        serverEvent(u"RESTARTED"_s, true, msg);
+        qCDebug(PMKCM) << "CUPS SERVER RE-STARTED" << msg;
+        Q_EMIT serverStarted();
     });
 
     qmlRegisterUncreatableMetaObject(PMTypes::staticMetaObject,
@@ -70,16 +82,30 @@ PrinterManager::PrinterManager(QObject *parent, const KPluginMetaData &metaData)
                                      0,
                                      "PPDType", // QML qualifier
                                      u"Error: for only enums"_s);
+
+    qDBusRegisterMetaType<DriverMatch>();
+    qDBusRegisterMetaType<DriverMatchList>();
+
+#ifdef SCP_INSTALL
+    qmlRegisterType<SCPInstaller>("org.kde.plasma.printmanager", 1, 0, "SCPInstaller");
+#endif
 }
 
-void PrinterManager::serverEvent(const QString &event, bool reload, const QString &msg)
+void PrinterManager::initOSRelease()
 {
-    qCWarning(PMKCM) << "SERVER" << event << msg << reload;
-    if (reload) {
-        QTimer::singleShot(500, this, &PrinterManager::getServerSettings);
-    } else {
-        m_serverSettingsLoaded = false;
-    }
+    KOSRelease os;
+    m_osName = os.name();
+    m_osBugReportUrl = os.bugReportUrl();
+}
+
+QString PrinterManager::osName() const
+{
+    return m_osName;
+}
+
+QString PrinterManager::osBugReportUrl() const
+{
+    return m_osBugReportUrl;
 }
 
 void PrinterManager::getRemotePrinters(const QString &uri, const QString &uriScheme)
@@ -154,29 +180,12 @@ void PrinterManager::clearRecommendedDrivers()
     m_recommendedDrivers.clear();
 }
 
-void PrinterManager::getDriversFinished(const QDBusMessage &message)
+bool PrinterManager::isSCPAvailable()
 {
-    if (message.type() == QDBusMessage::ReplyMessage && message.arguments().size() == 1) {
-        const auto args = message.arguments();
-        const auto arg = args.first().value<QDBusArgument>();
-        const DriverMatchList driverMatchList = qdbus_cast<DriverMatchList>(arg);
-        for (const DriverMatch &driverMatch : driverMatchList) {
-            if (driverMatch.match == u"none"_s) {
-                continue;
-            }
-
-            m_recommendedDrivers.append(QVariantMap({{u"match"_s, driverMatch.match}, {u"ppd-name"_s, driverMatch.ppd}, {u"ppd-type"_s, PMTypes::Auto}}));
-        }
-    } else {
-        qCWarning(PMKCM) << "Unexpected message" << message;
-    }
-    Q_EMIT recommendedDriversLoaded();
-}
-
-void PrinterManager::getDriversFailed(const QDBusError &error, const QDBusMessage &message)
-{
-    qCWarning(PMKCM) << "Failed to get best drivers" << error << message;
-    Q_EMIT recommendedDriversLoaded();
+    // The service is activatable, so check the list
+    const auto conn = QDBusConnection::sessionBus().interface();
+    const auto list = conn->activatableServiceNames();
+    return list.value().contains("org.fedoraproject.Config.Printing"_L1);
 }
 
 void PrinterManager::savePrinter(const QString &name, const QVariantMap &saveArgs, bool isClass)
@@ -192,16 +201,29 @@ void PrinterManager::savePrinter(const QString &name, const QVariantMap &saveArg
         }
     }
 
-    bool isDefault = args.take(u"isDefault"_s).toBool();
-    bool autoConfig = args.take(u"autoConfig"_s).toBool();
-    // add mode
-    if (args.take(u"add"_s).toBool()) {
+    const bool addMode = args.take(u"add"_s).toBool();
+    // Will only be set if default is changed to true
+    const bool isDefault = args.take(u"isDefault"_s).toBool();
+
+    if (addMode) {
         args[KCUPS_PRINTER_STATE] = IPP_PRINTER_IDLE;
     }
 
-    qCWarning(PMKCM) << "Saving printer settings" << Qt::endl << name << fileName << Qt::endl << args;
+    qCDebug(PMKCM) << (addMode ? "New Printer:" : "Change Printer:") << name << "isClass?" << isClass << "Changing Default?" << isDefault << "filename"
+                   << fileName << Qt::endl
+                   << "args" << args;
 
     QPointer<KCupsRequest> request = new KCupsRequest;
+
+    /** Default printer is handled by CUPS independently of the other printer
+     * attributes. If changing an existing printer and "default" is set
+     * then save immediately
+     */
+    if (!addMode && isDefault) {
+        request->setDefaultPrinter(name);
+        request->waitTillFinished();
+    }
+
     if (isClass) {
         // Member list is a QVariantList, kcupslib wants to see
         // a QStringList
@@ -209,21 +231,22 @@ void PrinterManager::savePrinter(const QString &name, const QVariantMap &saveArg
         if (!list.value<QVariantList>().empty()) {
             args.insert(KCUPS_MEMBER_URIS, list.toStringList());
         }
-        request->addOrModifyClass(name, args);
+        if (!args.isEmpty()) {
+            request->addOrModifyClass(name, args);
+            request->waitTillFinished();
+        }
     } else {
-        request->addOrModifyPrinter(name, args, fileName);
+        if (!args.isEmpty()) {
+            request->addOrModifyPrinter(name, args, fileName);
+            request->waitTillFinished();
+        }
     }
-    request->waitTillFinished();
 
-    // Printer isDefault is exclusive
-    if (isDefault) {
+    // If adding a new printer and it's set as default, save explicitly
+    if (addMode && isDefault) {
         request->setDefaultPrinter(name);
         request->waitTillFinished();
     }
-    if (autoConfig) {
-        request->printCommand(name, u"AutoConfigure"_s, i18n("Set Default Options"));
-        request->waitTillFinished();
-    }
 
     if (request) {
         if (!request->hasError()) {
@@ -292,7 +315,7 @@ QVariantMap PrinterManager::getPrinterPPD(const QString &name)
         codec = QStringDecoder(QStringDecoder::Utf8);
     }
 
-    qCWarning(PMKCM) << codec(ppd->pcfilename) << codec(ppd->modelname) << codec(ppd->shortnickname);
+    qCDebug(PMKCM) << codec(ppd->pcfilename) << codec(ppd->modelname) << codec(ppd->shortnickname);
 
     QString make, makeAndModel, file;
     if (ppd->manufacturer) {
@@ -341,20 +364,54 @@ void PrinterManager::resumePrinter(const QString &name)
     request->resumePrinter(name);
 }
 
+bool PrinterManager::isIPPCapable(const QString &uri)
+{
+    /**
+     * Per CUPS
+     * "dnssd:" URIs with "._ipp._tcp" or "._ipps._tcp"
+     *  or "ipp:" or "ipps:" URIs can be of a driverless printer
+    */
+    if (uri.startsWith(u"dnssd:"_s, Qt::CaseInsensitive)
+        && (uri.contains(u"._ipp._tcp"_s, Qt::CaseInsensitive)
+            || uri.contains(u"._ipps._tcp"_s, Qt::CaseInsensitive))) {
+        return true;
+    }
+
+    if (uri.startsWith(u"ipp:"_s, Qt::CaseInsensitive)
+            || uri.startsWith(u"ipps:"_s, Qt::CaseInsensitive)) {
+        return true;
+    }
+
+    return false;
+}
+
 void PrinterManager::getRecommendedDrivers(const QString &deviceId, const QString &makeAndModel, const QString &deviceUri)
 {
-    qCDebug(PMKCM) << deviceId << makeAndModel << deviceUri;
+    qCDebug(PMKCM) << "getRecommendedDrivers for:" << makeAndModel << deviceUri << deviceId;
 
     m_recommendedDrivers.clear();
-    QDBusMessage message;
-    message = QDBusMessage::createMethodCall(u"org.fedoraproject.Config.Printing"_s,
-                                             u"/org/fedoraproject/Config/Printing"_s,
-                                             u"org.fedoraproject.Config.Printing"_s,
-                                             u"GetBestDrivers"_s);
-    message << deviceId;
-    message << makeAndModel;
-    message << deviceUri;
-    QDBusConnection::sessionBus().callWithCallback(message, this, SLOT(getDriversFinished(QDBusMessage)), SLOT(getDriversFailed(QDBusError, QDBusMessage)));
+
+    auto call = QDBusMessage::createMethodCall(u"org.fedoraproject.Config.Printing"_s, u"/org/fedoraproject/Config/Printing"_s, u"org.fedoraproject.Config.Printing"_s, u"GetBestDrivers"_s);
+    call.setArguments({deviceId, makeAndModel, deviceUri});
+    const auto pending = QDBusConnection::sessionBus().asyncCall(call);
+    const auto watcher = new QDBusPendingCallWatcher(pending, this);
+
+    QObject::connect(watcher, &QDBusPendingCallWatcher::finished, this, [this](QDBusPendingCallWatcher *w) {
+        QDBusPendingReply<DriverMatchList> reply(*w);
+        if (reply.isError()) {
+            qCWarning(PMKCM) << "Failed to get best drivers" << reply.error().message();
+        } else {
+            const auto dml = reply.value();
+            for (const auto &driverMatch : dml) {
+                if (driverMatch.match == u"none"_s) {
+                    continue;
+                }
+                m_recommendedDrivers.append(QVariantMap({{u"match"_s, driverMatch.match}, {u"ppd-name"_s, driverMatch.ppd}, {u"ppd-type"_s, PMTypes::Auto}}));
+            }
+        }
+        Q_EMIT recommendedDriversLoaded();
+        w->deleteLater();
+    });
 }
 
 KCupsRequest *PrinterManager::setupRequest(std::function<void()> finished)
@@ -440,24 +497,26 @@ void PrinterManager::saveServerSettings(const QVariantMap &settings)
     server.setAllowRemoteAdmin(settings.value(QLatin1String(CUPS_SERVER_REMOTE_ADMIN), false).toBool());
     server.setAllowPrintingFromInternet(settings.value(QLatin1String(CUPS_SERVER_REMOTE_ANY), false).toBool());
 
+    /**
+     * The CUPS server will stop/start when settings change.  This will either be a stop signal
+     * from CUPS and/or the non-fatal error below
+    */
     auto request = new KCupsRequest;
     request->setServerSettings(server);
     request->waitTillFinished();
 
-    if (request->hasError()) {
-        if (request->error() == IPP_SERVICE_UNAVAILABLE || request->error() == IPP_INTERNAL_ERROR || request->error() == IPP_AUTHENTICATION_CANCELED) {
-            qCWarning(PMKCM) << "Server error:" << request->serverError() << request->errorMsg();
-            // Server is restarting, or auth was canceled, update the settings in one second
-            //            QTimer::singleShot(1000, this, &PrinterManager::getServerSettings);
-        } else {
-            // Force the settings to be retrieved again
-            getServerSettings();
-            Q_EMIT requestError(i18nc("@info", "Server error: (%1): %2", request->serverError(), request->errorMsg()));
-        }
-    } else {
-        qCWarning(PMKCM) << "SERVER SETTINGS SET!" << settings;
+    if (request->error() == IPP_AUTHENTICATION_CANCELED) {
+        // If the user has cancelled auth
+        Q_EMIT requestError(i18nc("@info", "Server Settings Not Saved: (%1): %2", request->serverError(), request->errorMsg()));
+    } else if (request->error() == IPP_SERVICE_UNAVAILABLE || request->error() == IPP_INTERNAL_ERROR) {
+        // this is the expected "error" after settings have changed as the server is restarting
         m_serverSettings = settings;
+        Q_EMIT serverSettingsChanged();
+        qCDebug(PMKCM) << "CUPS SETTINGS SAVED!" << settings;
+    } else {
+        Q_EMIT requestError(i18nc("@info", "Fatal Save Server Settings: (%1): %2", request->serverError(), request->errorMsg()));
     }
+
     request->deleteLater();
 }
 
